diff --git a/src/backend/commands/copy.c b/src/backend/commands/copy.c
index ef301b6328..ef1c5337f5 100644
--- a/src/backend/commands/copy.c
+++ b/src/backend/commands/copy.c
@@ -2803,12 +2803,13 @@ CopyFrom(CopyState cstate)
 			 * tuples inserted by an INSERT command.
 			 */
 			processed++;
+		}
 
-			if (saved_resultRelInfo)
-			{
-				resultRelInfo = saved_resultRelInfo;
-				estate->es_result_relation_info = resultRelInfo;
-			}
+		/* Restore the saved ResultRelInfo */
+		if (saved_resultRelInfo)
+		{
+			resultRelInfo = saved_resultRelInfo;
+			estate->es_result_relation_info = resultRelInfo;
 		}
 	}
 
diff --git a/src/backend/executor/nodeModifyTable.c b/src/backend/executor/nodeModifyTable.c
index 4742593374..d3c459282c 100644
--- a/src/backend/executor/nodeModifyTable.c
+++ b/src/backend/executor/nodeModifyTable.c
@@ -259,7 +259,6 @@ ExecInsert(ModifyTableState *mtstate,
 {
 	HeapTuple	tuple;
 	ResultRelInfo *resultRelInfo;
-	ResultRelInfo *saved_resultRelInfo = NULL;
 	Relation	resultRelationDesc;
 	Oid			newId;
 	List	   *recheckIndexes = NIL;
@@ -275,100 +274,6 @@ ExecInsert(ModifyTableState *mtstate,
 	 * get information on the (current) result relation
 	 */
 	resultRelInfo = estate->es_result_relation_info;
-
-	/* Determine the partition to heap_insert the tuple into */
-	if (mtstate->mt_partition_dispatch_info)
-	{
-		int			leaf_part_index;
-		TupleConversionMap *map;
-
-		/*
-		 * Away we go ... If we end up not finding a partition after all,
-		 * ExecFindPartition() does not return and errors out instead.
-		 * Otherwise, the returned value is to be used as an index into arrays
-		 * mt_partitions[] and mt_partition_tupconv_maps[] that will get us
-		 * the ResultRelInfo and TupleConversionMap for the partition,
-		 * respectively.
-		 */
-		leaf_part_index = ExecFindPartition(resultRelInfo,
-											mtstate->mt_partition_dispatch_info,
-											slot,
-											estate);
-		Assert(leaf_part_index >= 0 &&
-			   leaf_part_index < mtstate->mt_num_partitions);
-
-		/*
-		 * Save the old ResultRelInfo and switch to the one corresponding to
-		 * the selected partition.
-		 */
-		saved_resultRelInfo = resultRelInfo;
-		resultRelInfo = mtstate->mt_partitions + leaf_part_index;
-
-		/* We do not yet have a way to insert into a foreign partition */
-		if (resultRelInfo->ri_FdwRoutine)
-			ereport(ERROR,
-					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-					 errmsg("cannot route inserted tuples to a foreign table")));
-
-		/* For ExecInsertIndexTuples() to work on the partition's indexes */
-		estate->es_result_relation_info = resultRelInfo;
-
-		/*
-		 * If we're capturing transition tuples, we might need to convert from
-		 * the partition rowtype to parent rowtype.
-		 */
-		if (mtstate->mt_transition_capture != NULL)
-		{
-			if (resultRelInfo->ri_TrigDesc &&
-				(resultRelInfo->ri_TrigDesc->trig_insert_before_row ||
-				 resultRelInfo->ri_TrigDesc->trig_insert_instead_row))
-			{
-				/*
-				 * If there are any BEFORE or INSTEAD triggers on the
-				 * partition, we'll have to be ready to convert their result
-				 * back to tuplestore format.
-				 */
-				mtstate->mt_transition_capture->tcs_original_insert_tuple = NULL;
-				mtstate->mt_transition_capture->tcs_map =
-					mtstate->mt_transition_tupconv_maps[leaf_part_index];
-			}
-			else
-			{
-				/*
-				 * Otherwise, just remember the original unconverted tuple, to
-				 * avoid a needless round trip conversion.
-				 */
-				mtstate->mt_transition_capture->tcs_original_insert_tuple = tuple;
-				mtstate->mt_transition_capture->tcs_map = NULL;
-			}
-		}
-		if (mtstate->mt_oc_transition_capture != NULL)
-			mtstate->mt_oc_transition_capture->tcs_map =
-				mtstate->mt_transition_tupconv_maps[leaf_part_index];
-
-		/*
-		 * We might need to convert from the parent rowtype to the partition
-		 * rowtype.
-		 */
-		map = mtstate->mt_partition_tupconv_maps[leaf_part_index];
-		if (map)
-		{
-			Relation	partrel = resultRelInfo->ri_RelationDesc;
-
-			tuple = do_convert_tuple(tuple, map);
-
-			/*
-			 * We must use the partition's tuple descriptor from this point
-			 * on, until we're finished dealing with the partition. Use the
-			 * dedicated slot for that.
-			 */
-			slot = mtstate->mt_partition_tuple_slot;
-			Assert(slot != NULL);
-			ExecSetSlotDescriptor(slot, RelationGetDescr(partrel));
-			ExecStoreTuple(tuple, slot, InvalidBuffer, true);
-		}
-	}
-
 	resultRelationDesc = resultRelInfo->ri_RelationDesc;
 
 	/*
@@ -477,7 +382,7 @@ ExecInsert(ModifyTableState *mtstate,
 		 * No need though if the tuple has been routed, and a BR trigger
 		 * doesn't exist.
 		 */
-		if (saved_resultRelInfo != NULL &&
+		if (resultRelInfo->ri_PartitionRoot != NULL &&
 			!(resultRelInfo->ri_TrigDesc &&
 			  resultRelInfo->ri_TrigDesc->trig_insert_before_row))
 			check_partition_constr = false;
@@ -645,9 +550,6 @@ ExecInsert(ModifyTableState *mtstate,
 	if (resultRelInfo->ri_projectReturning)
 		result = ExecProcessReturning(resultRelInfo, slot, planSlot);
 
-	if (saved_resultRelInfo)
-		estate->es_result_relation_info = saved_resultRelInfo;
-
 	return result;
 }
 
@@ -1378,6 +1280,110 @@ ExecOnConflictUpdate(ModifyTableState *mtstate,
 	return true;
 }
 
+/*
+ * ExecPrepareTupleRouting --- prepare for tuple routing of one tuple
+ */
+static TupleTableSlot *
+ExecPrepareTupleRouting(ModifyTableState *mtstate,
+						EState *estate,
+						ResultRelInfo *rootRelInfo,
+						TupleTableSlot *slot)
+{
+	ResultRelInfo *resultRelInfo;
+	HeapTuple	tuple;
+	int			leaf_part_index;
+	TupleConversionMap *map;
+
+	/*
+	 * Away we go ... If we end up not finding a partition after all,
+	 * ExecFindPartition() does not return and errors out instead.
+	 * Otherwise, the returned value is to be used as an index into arrays
+	 * mtstate->mt_partitions[] and mtstate->mt_partition_tupconv_maps[] that
+	 * will get us the ResultRelInfo and TupleConversionMap for the partition,
+	 * respectively.
+	 */
+	leaf_part_index = ExecFindPartition(rootRelInfo,
+										mtstate->mt_partition_dispatch_info,
+										slot,
+										estate);
+	Assert(leaf_part_index >= 0 &&
+		   leaf_part_index < mtstate->mt_num_partitions);
+
+	/* Get the ResultRelInfo corresponding to the selected partition. */
+	resultRelInfo = &mtstate->mt_partitions[leaf_part_index];
+
+	/* We do not yet have a way to insert into a foreign partition */
+	if (resultRelInfo->ri_FdwRoutine)
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("cannot route inserted tuples to a foreign table")));
+
+	/*
+	 * For ExecInsert(), make it look like we are inserting into the partition.
+	 */
+	estate->es_result_relation_info = resultRelInfo;
+
+	/* Get the heap tuple out of the tuple table slot. */
+	tuple = ExecMaterializeSlot(slot);
+
+	/*
+	 * If we're capturing transition tuples, we might need to convert from the
+	 * partition rowtype to parent rowtype.
+	 */
+	if (mtstate->mt_transition_capture != NULL)
+	{
+		if (resultRelInfo->ri_TrigDesc &&
+			(resultRelInfo->ri_TrigDesc->trig_insert_before_row ||
+			 resultRelInfo->ri_TrigDesc->trig_insert_instead_row))
+		{
+			/*
+			 * If there are any BEFORE or INSTEAD triggers on the partition,
+			 * we'll have to be ready to convert their result back to
+			 * tuplestore format.
+			 */
+			mtstate->mt_transition_capture->tcs_original_insert_tuple = NULL;
+			mtstate->mt_transition_capture->tcs_map =
+				mtstate->mt_transition_tupconv_maps[leaf_part_index];
+		}
+		else
+		{
+			/*
+			 * Otherwise, just remember the original unconverted tuple, to
+			 * avoid a needless round trip conversion.
+			 */
+			mtstate->mt_transition_capture->tcs_original_insert_tuple = tuple;
+			mtstate->mt_transition_capture->tcs_map = NULL;
+		}
+	}
+	if (mtstate->mt_oc_transition_capture != NULL)
+		mtstate->mt_oc_transition_capture->tcs_map =
+			mtstate->mt_transition_tupconv_maps[leaf_part_index];
+
+	/*
+	 * We might need to convert from the parent rowtype to the partition
+	 * rowtype.
+	 */
+	map = mtstate->mt_partition_tupconv_maps[leaf_part_index];
+	if (map)
+	{
+		Relation	partrel = resultRelInfo->ri_RelationDesc;
+
+		tuple = do_convert_tuple(tuple, map);
+
+		/*
+		 * We must use the partition's tuple descriptor from this point
+		 * on, until we're finished dealing with the partition. Use the
+		 * dedicated slot for that.
+		 */
+		slot = mtstate->mt_partition_tuple_slot;
+		Assert(slot != NULL);
+		ExecSetSlotDescriptor(slot, RelationGetDescr(partrel));
+		ExecStoreTuple(tuple, slot, InvalidBuffer, true);
+	}
+
+	return slot;
+}
+
 
 /*
  * Process BEFORE EACH STATEMENT triggers
@@ -1545,6 +1551,7 @@ ExecSetupTransitionCaptureState(ModifyTableState *mtstate, EState *estate)
 	}
 }
 
+
 /* ----------------------------------------------------------------
  *	   ExecModifyTable
  *
@@ -1763,9 +1770,20 @@ ExecModifyTable(PlanState *pstate)
 		switch (operation)
 		{
 			case CMD_INSERT:
+
+				/* Prepare for tuple routing of the tuple if needed */
+				if (node->mt_partition_dispatch_info)
+					slot = ExecPrepareTupleRouting(node, estate,
+												   resultRelInfo, slot);
 				slot = ExecInsert(node, slot, planSlot,
 								  node->mt_arbiterindexes, node->mt_onconflict,
 								  estate, node->canSetTag);
+				/*
+				 * Revert back the active result relation that
+				 * ExecPrepareTupleRouting would have changed.
+				 */
+				if (node->mt_partition_dispatch_info)
+					estate->es_result_relation_info = resultRelInfo;
 				break;
 			case CMD_UPDATE:
 				slot = ExecUpdate(node, tupleid, oldtuple, slot, planSlot,
